#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <sys/wait.h>

#include <glib.h>
#include <gps.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus-glib.h>


#define	DEFAULT_TAILTIME	300
#define SERIAL_POLL_INTERVAL	100
#define TRACKLIMIT		5

static struct {
	char	*serialdevice;
	int	tailtime;
	char	*destination;

	char	*recordcmd;
	char	*playcmd;
} config;

struct gpsdata_s {
	time_t  time;
	gint32  mode;
	gdouble ept;
	gdouble latitude;
	gdouble longitude;
	gdouble eph;
	gdouble altitude;
	gdouble epv;
	gdouble track;
	gdouble epd;
	gdouble speed;
	gdouble eps;
	gdouble climb;
	gdouble epc;
};


static struct {
	time_t			starttime;
	char			starttimestr[64];

	int			serialfd;
	int			laststatus;

	struct gps_data_t	*gpsdata;
	pthread_t		gpsthread;

	GString			*trackdir;
	GMainLoop		*mainloop;

	struct {
		time_t			updated;
		time_t  		old_time;
		struct gpsdata_s	data;
	} gps;

	struct {
		pid_t		pid;
		GString		*filename;
		GString		*filepath;
		guint		tosrcid;
	} rec;

	struct {
		pid_t		pid;
	} replay;

	struct {
		int		trkfd;
		int		wptfd;

		int		trackpoint;
		int		waypoint;

		GString		*str;

		int		first;
		int		intrack;
	} gpx;
} g;

static void gpx_write_waypoint(char *filename);

static void child_stop(pid_t pid) {
	printf("Killing pid %d\n", pid);
	kill(pid, SIGTERM);
}

static GPid child_start(char *cmdline) {
	gint		argcp;
	gchar		**argvp;
	GError		*error=NULL;
	gboolean	result;
	GPid		cpid;

	result=g_shell_parse_argv(cmdline, &argcp, &argvp, &error);
	if (!result) {
		printf("shell parsing failed: %s\n", error->message);
		exit(-1);
	}
	result=g_spawn_async(NULL, argvp, NULL, 0, NULL, NULL, &cpid, &error);
	if (!result) {
		printf("spawning command async failed: %s\n", error->message);
		exit(-1);
	}
	g_strfreev(argvp);

	return cpid;
}

static void record_init(void ) {
	g.rec.filename=g_string_sized_new(256);
	g.rec.filepath=g_string_sized_new(256);
}

static void record_start(void ) {
	gchar	*cmdline;

	/* Create filename in global buffer */
	g_string_printf(g.rec.filename, "audiowaypoint-%lu-%f-%f-%d.wav", 
		g.gps.data.time,
		g.gps.data.latitude,
		g.gps.data.longitude,
		++g.gpx.waypoint);

	printf("starting record to: %s\n", g.rec.filename->str);

	g_string_printf(g.rec.filepath, "%s/%s", g.trackdir->str, 
		g.rec.filename->str);

	gpx_write_waypoint(g.rec.filename->str);

	/* Start recording as child */
	cmdline=g_strjoin(" ", config.recordcmd, g.rec.filepath->str, NULL);
	g.rec.pid=child_start(cmdline);
	free(cmdline);
}

static int record_running(void ) {
	return (g.rec.pid != 0);
}

static void record_stop(void ) {
	if (record_running()) {
		child_stop(g.rec.pid);
		g.rec.pid=0;
	}
	printf("Stopped recording\n");
}

static void replay_start(void ) {
	char	*cmdline;

	cmdline=g_strjoin(" ", config.playcmd, g.rec.filepath->str, NULL);
	g.replay.pid=child_start(cmdline);
	free(cmdline);
}

static gboolean record_stop_timeout(gpointer data) {
	record_stop();
	replay_start();

	return FALSE;
}

static void record_trigger(int	status) {
	if (status) {
		if (!record_running())
			record_start();

		/* If we were running before remove the tail timer */
		if (g.rec.tosrcid) {
			g_source_remove(g.rec.tosrcid);
			g.rec.tosrcid=0;
		}
	} else {
		/* If we were running while trigger release add the tailtime */
		if (record_running()) 
			g.rec.tosrcid=g_timeout_add(config.tailtime, record_stop_timeout, NULL);
	}
}

int serial_open(char *name) {
	int		fd, res;
	int		status;
	struct termios	tios;

	fd=open(config.serialdevice, O_RDWR|O_EXCL|O_NONBLOCK|O_NOCTTY);

	if (fd < 0) {
		fprintf(stderr, "Failed to open %s as trigger dev\n",
			config.serialdevice);
		exit(-1);
	}

	memset(&tios, 0, sizeof(struct termios));
	tios.c_cflag=CS8|CREAD|CLOCAL|CSTOPB|CRTSCTS;
	res=tcsetattr(fd, TCSANOW, &tios);

	/* Assert DTR */
	res=ioctl(fd, TIOCMGET, &status);
	status|=TIOCM_DTR;
	ioctl(fd, TIOCMSET, &status);

	return fd;
}

static int serial_pollstatus(int fd) {
	int	status, res;

	res=ioctl(fd, TIOCMGET, &status);

	return status & TIOCM_DSR;
}

static gboolean serial_status_timer(gpointer data) {
	int	status;

	status=serial_pollstatus(g.serialfd);

	if (status != g.laststatus)
		record_trigger(status);

	g.laststatus=status;

	return TRUE;
}

/*
 *
 *  GPS Code
 *
 */

static void gpx_track_open(void ) {
	g_string_printf(g.gpx.str, "<trk><trkseg>\n");

	/* FIXME check for write success */
	write(g.gpx.trkfd, g.gpx.str->str, g.gpx.str->len);

	g.gpx.intrack=1;
}

static void gpx_track_close(void ) {
	const char	*eotrack="</trkseg></trk>\n";

	/* FIXME check for write success */
	write(g.gpx.trkfd, eotrack, strlen(eotrack));

	g.gpx.intrack=1;
}


static void gpx_write_waypoint(char *filename) {
	GString		*wp;
	struct tm	time;

	wp=g_string_sized_new(256);

	gmtime_r(&(g.gps.data.time), &time);

	g_string_append_printf(wp, "<wpt lat=\"%f\" lon=\"%f\">\n",
			g.gps.data.latitude, g.gps.data.longitude);
	if (!isnan(g.gps.data.altitude))
		g_string_append_printf(wp, "\t<ele>%f</ele>\n", g.gps.data.altitude);
	g_string_append_printf(wp, "\t<name>WP%04d</name>\n", g.gpx.waypoint);
	g_string_append_printf(wp, "\t<time>%04d-%02d-%02dT%02d:%02d:%02dZ</time>\n",
				time.tm_year+1900, time.tm_mon+1, time.tm_mday,
				time.tm_hour, time.tm_min, time.tm_sec);
	if(g.gps.data.mode==1)
		g_string_append_printf(wp, "\t<fix>none</fix>\n");
	else
		g_string_append_printf(wp, "\t<fix>%dd</fix>\n", g.gps.data.mode);
		
	g_string_append_printf(wp, "\t<link href=\"%s\">\n", filename);
	g_string_append_printf(wp, "\t\t<text>%s</text>\n", filename);
	g_string_append_printf(wp, "\t\t<type>audio/wav</type>\n");
	g_string_append_printf(wp, "\t</link>\n");
	g_string_append_printf(wp, "</wpt>\n");

	/* FIXME - check for write success */
	write(g.gpx.wptfd, wp->str, wp->len);

	g_string_free(wp, true);
}


static void gpx_write_track(void ) {	
	/* 
	 * we have a fix there - log the point
	 */
	if ((g.gps.data.time!=g.gps.old_time)&&g.gps.data.mode>1) {
		struct tm 	time;

		/* Make new track if the jump in time is above
		 * tracklimit.  Handle jumps both forward and
		 * backwards in time.  The clock sometimes jump
		 * backward when gpsd is submitting junk on the
		 * dbus. */
		if (fabs(g.gps.data.time - g.gps.old_time) > TRACKLIMIT && !g.gpx.first) {
			printf("%lu %lu\n", g.gps.data.time, g.gps.old_time);
			gpx_track_close();
			g.gpx.intrack = 0;
		}

		if (!g.gpx.intrack) {
			gpx_track_open();
			g.gpx.intrack = 1;
			if (g.gpx.first)
			    g.gpx.first = 0;
		}
		
		g.gps.old_time = g.gps.data.time;

		g_string_printf(g.gpx.str, "   <trkpt lat=\"%f\" lon=\"%f\">\n",
			g.gps.data.latitude, g.gps.data.longitude);
		if (!isnan(g.gps.data.altitude)) 
			g_string_append_printf(g.gpx.str, "    <ele>%f</ele>\n",
				g.gps.data.altitude);
		gmtime_r(&(g.gps.data.time), &time);
		g_string_append_printf(g.gpx.str, "    <time>%04d-%02d-%02dT%02d:%02d:%02dZ</time>\n",
					time.tm_year+1900, time.tm_mon+1, time.tm_mday,
					time.tm_hour, time.tm_min, time.tm_sec);
		if (g.gps.data.mode==1)
			g_string_append_printf(g.gpx.str, "    <fix>none</fix>\n");
		else
			g_string_append_printf(g.gpx.str, "    <fix>%dd</fix>\n", g.gps.data.mode);
		g_string_append_printf(g.gpx.str, "   </trkpt>\n");

		write(g.gpx.trkfd, g.gpx.str->str, g.gpx.str->len);
	}
}

static void gpx_write_prolog(int fd) {
	GString		*p;

	p=g_string_sized_new(256);

	g_string_append_printf(p, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	g_string_append_printf(p, "<gpx version=\"1.0\" creator=\"triggerrec\"\n");
	g_string_append_printf(p, "\txmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n");
	g_string_append_printf(p, "\txmlns=\"http://www.topografix.com/GPX/1/0\"\n");
	g_string_append_printf(p, "\txsi:schemaLocation=\"http://www.topografix.com/GPX/1/0 http://www.topografix.com/GPX/1/0/gpx.xsd\">\n");

	/* FIXME */
	write(fd, p->str, p->len);

	g_string_free(p, true);
}

static void gpx_write_epilog(int fd) {
	const char	*epi="\n</gpx>\n";
	write(fd, epi, strlen(epi));
}

static DBusHandlerResult gps_handle_message(DBusMessage* message) {
	DBusError		error;
	int			rc;
	double			ttime;
	struct gpsdata_s	data;

        dbus_error_init(&error);

        rc=dbus_message_get_args (message,
                               &error,
                               DBUS_TYPE_DOUBLE, &ttime,
                               DBUS_TYPE_INT32,  &data.mode,
                               DBUS_TYPE_DOUBLE, &data.ept,
                               DBUS_TYPE_DOUBLE, &data.latitude,
                               DBUS_TYPE_DOUBLE, &data.longitude,
                               DBUS_TYPE_DOUBLE, &data.eph,
                               DBUS_TYPE_DOUBLE, &data.altitude,
                               DBUS_TYPE_DOUBLE, &data.epv,
                               DBUS_TYPE_DOUBLE, &data.track,
                               DBUS_TYPE_DOUBLE, &data.epd,
                               DBUS_TYPE_DOUBLE, &data.speed,
                               DBUS_TYPE_DOUBLE, &data.eps,
                               DBUS_TYPE_DOUBLE, &data.climb,
                               DBUS_TYPE_DOUBLE, &data.epc,
                               DBUS_TYPE_INVALID);
        data.time = floor(ttime);

	/*
	 * For every valid gps position from gpsd i get a broken one - fix set
	 * but longitude and latitude are nan and time is negative (1901)
	 */
	if (rc && data.time > 0 && !isnan(data.latitude) && !isnan(data.longitude)) {
		memcpy(&g.gps.data, &data, sizeof(struct gpsdata_s));
		g.gps.updated=time(NULL);
		gpx_write_track();
	}

	printf("dbus_message_get_args %d %lu %s mode %d lat %f lon %f\n",
		rc, data.time,
		dbus_message_get_sender(message),
		data.mode, data.latitude, data.longitude);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult gps_signal_handler(DBusConnection *connection, DBusMessage *message) {
	if (dbus_message_is_signal(message, "org.gpsd", "fix")) {
		return gps_handle_message(message);
		
	}
	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;	
}

static void gps_init(void ) {
	DBusError	error;
	dbus_error_init(&error);
	DBusConnection *connection;

	connection = dbus_bus_get(DBUS_BUS_SYSTEM, &error);
	if(dbus_error_is_set(&error)) {
		fprintf(stderr, "Could not connect to DBUS for gps messages\n");
		exit(-1);
	}
	dbus_bus_add_match(connection, "type='signal'", &error);
	if(dbus_error_is_set(&error)) {
		fprintf(stderr, "Could not add match to DBUS connection for gps messages\n");
		exit(-1);
	}	
	if(!dbus_connection_add_filter(connection, (DBusHandleMessageFunction) gps_signal_handler, NULL, NULL)) {
		fprintf(stderr, "Could not add filter to DBUS connection for gps messages\n");
		exit(-1);
	}

	dbus_connection_setup_with_g_main(connection, NULL);

#if 0
	g.gpsdata=gps_open("localhost", "2947");

	if (!g.gpsdata) {
		fprintf(stderr, "Error opening gpsd\n");
		exit(-1);
	}
	
	gps_set_callback(g.gpsdata, gps_callback, &g.gpsthread);
#endif
}

static void serial_status_timerinit(void ) {
	g_timeout_add(SERIAL_POLL_INTERVAL, serial_status_timer, NULL);
}

void trackdir_create(void ) {
	struct tm	*ltime;
	int		res;

	g.starttime=time(NULL);
	ltime=localtime(&g.starttime);

	strftime(g.starttimestr, sizeof(g.starttimestr),
		"%Y%m%d-%H%M%S", ltime);

	g.trackdir=g_string_sized_new(256);
	g_string_printf(g.trackdir, "%s/audiotrack-%s", config.destination, g.starttimestr);

	res=mkdir(g.trackdir->str, 0755);

	if (res && errno != EEXIST) {
		fprintf(stderr, "Failed to create destination dir %s\n", g.trackdir->str);
		exit(-1);
	}
}
static void sigint_handler(int sig) {

	if (g.gpx.intrack)
		gpx_track_close();
	gpx_write_epilog(g.gpx.trkfd);
	close(g.gpx.trkfd);

	gpx_write_epilog(g.gpx.wptfd);
	close(g.gpx.wptfd);

	exit(0);
}

static void gpx_init(void ) {
	gchar		*filename;

	/* Open trackfile */
	filename=g_strjoin(NULL, g.trackdir->str, "/track-", g.starttimestr, ".gpx", NULL);

	g.gpx.trkfd=open(filename, O_TRUNC|O_RDWR|O_CREAT, 0644);

	if (g.gpx.trkfd < 0) {
		printf("Failed to open track file %s\n", filename);
		exit(-1);
	}

	free(filename);

	gpx_write_prolog(g.gpx.trkfd);

	/* Open Waypoint file */
	filename=g_strjoin(NULL, g.trackdir->str, "/waypoint-", g.starttimestr, ".gpx", NULL);

	g.gpx.wptfd=open(filename, O_TRUNC|O_RDWR|O_CREAT, 0644);

	if (g.gpx.wptfd < 0) {
		printf("Failed to open waypoint file %s\n", filename);
		exit(-1);
	}

	free(filename);

	gpx_write_prolog(g.gpx.wptfd);

	g.gpx.str=g_string_sized_new(256);
	g.gpx.first=1;

	signal(SIGINT, sigint_handler);	
}

static void usage(void ) {
	printf("triggerrec Usage:\n");
	printf(" -s	- Serialdevice default: /dev/ttyS0:\n");
	printf(" -t	- tailtime in ms to record after trigger release\n");
	printf(" -d	- Destination directory\n");
	exit(0);
}

static void sigchld_handler(int signal) {
	int		status;
	pid_t		pid;

	pid=wait(&status);

	printf("pid %d terminated with status %08x\n", pid, status);
}


int main(int argc, char **argv) {
	extern char	*optarg;
	char		ch;

	config.serialdevice="/dev/ttyS0";
	config.tailtime=DEFAULT_TAILTIME;
	config.playcmd="/usr/bin/aplay -D plughw:1";
	config.recordcmd="/usr/bin/arecord -D plughw:1 -f S16_LE -r 44100";

	while((ch=getopt(argc, argv, "s:t:d:")) != -1) {
		switch(ch) {
			case 's':
				config.serialdevice=strdup(optarg);
				break;
			case 't':
				config.tailtime=strtol(optarg, NULL, 10);
				break;
			case 'd':
				config.destination=strdup(optarg);
				break;
			case 'r':
				config.recordcmd=strdup(optarg);
				break;
			case 'p':
				config.playcmd=strdup(optarg);
				break;
			default:
				usage();
				break;
		}
	
	}

	signal(SIGCHLD, sigchld_handler);

	g.mainloop=g_main_loop_new(NULL, FALSE);

	trackdir_create();

	g.serialfd=serial_open(config.serialdevice);
	serial_status_timerinit();

	gps_init();
	record_init();
	gpx_init();

	g_main_loop_run(g.mainloop);

	exit(0);
}
